@@ -19,7 +19,7 @@
  * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#ident "$Id: obj_ia64.c 1.8 Tue, 12 Sep 2000 14:01:20 +1100 kaos $"
+#ident "$Id: obj_ia64.c 1.9 Sat, 25 Nov 2000 19:29:38 +1100 kaos $"
 
 #include <string.h>
 #include <assert.h>
@@ -1010,14 +1010,15 @@ arch_init_module (struct obj_file *f, struct module *mod)
     ia64_file_t *ifile = (ia64_file_t *)f;
     Elf64_Addr *opd = (Elf64_Addr *)(ifile->opd->contents);
 
-    if ((*opd++ = mod->init) == 0)
-	return 0;
-    *opd++ = ifile->gp;
-    mod->init = ifile->opd->header.sh_addr;
-
-    if ((*opd++ = mod->cleanup) != 0)
+    if ((opd[0] = mod->init) != 0)
     {
-	*opd = ifile->gp;
+	opd[1] = ifile->gp;
+	mod->init = ifile->opd->header.sh_addr;
+    }
+
+    if ((opd[2] = mod->cleanup) != 0)
+    {
+	opd[3] = ifile->gp;
 	mod->cleanup = ifile->opd->header.sh_addr + 16;
     }
 
